def _OutputBuffer_InitAndGrow(buffer, ob, max_length):
    res, ob.dst = buffer.init_and_grow(max_length)
    ob.size = res
    ob.pos = 0


def _OutputBuffer_InitWithSize(buffer, ob, max_length, init_size):
    # Get block size
    if 0 <= max_length < init_size:
        block_size = max_length
    else:
        block_size = init_size

    res, ob.dst = buffer.init_with_size(block_size)

    # Set max_length, InitWithSize doesn't do this
    buffer.max_length = max_length
    ob.size = res
    ob.pos = 0


def _OutputBuffer_Grow(buffer, ob):
    res, ob.dst = buffer.grow(0)
    ob.size = res
    ob.pos = 0


def _OutputBuffer_Finish(buffer, ob):
    return buffer.finish(ob.size - ob.pos)


def _OutputBuffer_ReachedMaxLength(buffer, ob):
    return buffer.allocated == buffer.max_length



# if present, the C extension takes precedence over this file during import

from backports.zstd._cffi import (
    ZSTD_CLEVEL_DEFAULT,
    ZSTD_btlazy2,
    ZSTD_btopt,
    ZSTD_btultra,
    ZSTD_btultra2,
    ZSTD_c_chainLog,
    ZSTD_c_checksumFlag,
    ZSTD_c_compressionLevel,
    ZSTD_c_contentSizeFlag,
    ZSTD_c_dictIDFlag,
    ZSTD_c_enableLongDistanceMatching,
    ZSTD_c_hashLog,
    ZSTD_c_jobSize,
    ZSTD_c_ldmBucketSizeLog,
    ZSTD_c_ldmHashLog,
    ZSTD_c_ldmHashRateLog,
    ZSTD_c_ldmMinMatch,
    ZSTD_c_minMatch,
    ZSTD_c_nbWorkers,
    ZSTD_c_overlapLog,
    ZSTD_c_searchLog,
    ZSTD_c_strategy,
    ZSTD_c_targetLength,
    ZSTD_c_windowLog,
    ZSTD_d_windowLogMax,
    ZSTD_dfast,
    ZSTD_DStreamOutSize,
    ZSTD_fast,
    ZSTD_greedy,
    ZSTD_lazy,
    ZSTD_lazy2,
    ZstdCompressor,
    ZstdDecompressor,
    ZstdDict,
    ZstdError,
    finalize_dict,
    get_frame_info,
    get_frame_size,
    get_param_bounds,
    set_parameter_types,
    train_dict,
    zstd_version,
    zstd_version_number,
)



import enum

from backports.zstd._zstd_cffi import ffi as _ffi
from backports.zstd._zstd_cffi import lib as _lib

_new_nonzero = _ffi.new_allocator(should_clear_after_alloc=False)


class ZstdError(Exception):
    "An error occurred in the zstd library."


def _nbytes(dat, /):
    if isinstance(dat, (bytes, bytearray)):
        return len(dat)
    with memoryview(dat) as mv:
        return mv.nbytes


class _ErrorType(enum.IntEnum):
    ERR_DECOMPRESS = 0
    ERR_COMPRESS = enum.auto()
    ERR_SET_PLEDGED_INPUT_SIZE = enum.auto()
    ERR_LOAD_D_DICT = enum.auto()
    ERR_LOAD_C_DICT = enum.auto()
    ERR_GET_C_BOUNDS = enum.auto()
    ERR_GET_D_BOUNDS = enum.auto()
    ERR_SET_C_LEVEL = enum.auto()
    ERR_TRAIN_DICT = enum.auto()
    ERR_FINALIZE_DICT = enum.auto()


class _DictionaryType(enum.IntEnum):
    DICT_TYPE_DIGESTED = 0
    DICT_TYPE_UNDIGESTED = enum.auto()
    DICT_TYPE_PREFIX = enum.auto()


_ErrorTypeMessages = {
    _ErrorType.ERR_DECOMPRESS: "Unable to decompress Zstandard data: %s",
    _ErrorType.ERR_COMPRESS: "Unable to compress Zstandard data: %s",
    _ErrorType.ERR_SET_PLEDGED_INPUT_SIZE: "Unable to set pledged uncompressed content size: %s",
    _ErrorType.ERR_LOAD_D_DICT: "Unable to load Zstandard dictionary or prefix for decompression: %s",
    _ErrorType.ERR_LOAD_C_DICT: "Unable to load Zstandard dictionary or prefix for compression: %s",
    _ErrorType.ERR_GET_C_BOUNDS: "Unable to get zstd compression parameter bounds: %s",
    _ErrorType.ERR_GET_D_BOUNDS: "Unable to get zstd decompression parameter bounds: %s",
    _ErrorType.ERR_SET_C_LEVEL: "Unable to set zstd compression level: %s",
    _ErrorType.ERR_TRAIN_DICT: "Unable to train the Zstandard dictionary: %s",
    _ErrorType.ERR_FINALIZE_DICT: "Unable to finalize the Zstandard dictionary: %s",
}


def _set_zstd_error(type_, zstd_ret):
    raise ZstdError(
        _ErrorTypeMessages[type_]
        % _ffi.string(_lib.ZSTD_getErrorName(zstd_ret)).decode()
    )


_CpItems = {
    _lib.ZSTD_c_compressionLevel: "compression_level",
    _lib.ZSTD_c_windowLog: "window_log",
    _lib.ZSTD_c_hashLog: "hash_log",
    _lib.ZSTD_c_chainLog: "chain_log",
    _lib.ZSTD_c_searchLog: "search_log",
    _lib.ZSTD_c_minMatch: "min_match",
    _lib.ZSTD_c_targetLength: "target_length",
    _lib.ZSTD_c_strategy: "strategy",
    _lib.ZSTD_c_enableLongDistanceMatching: "enable_long_distance_matching",
    _lib.ZSTD_c_ldmHashLog: "ldm_hash_log",
    _lib.ZSTD_c_ldmMinMatch: "ldm_min_match",
    _lib.ZSTD_c_ldmBucketSizeLog: "ldm_bucket_size_log",
    _lib.ZSTD_c_ldmHashRateLog: "ldm_hash_rate_log",
    _lib.ZSTD_c_contentSizeFlag: "content_size_flag",
    _lib.ZSTD_c_checksumFlag: "checksum_flag",
    _lib.ZSTD_c_dictIDFlag: "dict_id_flag",
    _lib.ZSTD_c_nbWorkers: "nb_workers",
    _lib.ZSTD_c_jobSize: "job_size",
    _lib.ZSTD_c_overlapLog: "overlap_log",
}
_DpItems = {
    _lib.ZSTD_d_windowLogMax: "window_log_max",
}


def _set_parameter_error(is_compress, key, value):
    if is_compress:
        items = _CpItems
        type_ = "compression"
    else:
        items = _DpItems
        type_ = "decompression"

    # Find parameter's name
    try:
        name = items[key]
    # Unknown parameter
    except KeyError:
        name = "unknown parameter (key %d)" % key

    # Get parameter bounds
    try:
        if is_compress:
            bounds = _lib.ZSTD_cParam_getBounds(key)
        else:
            bounds = _lib.ZSTD_dParam_getBounds(key)
    except OverflowError:
        raise ValueError("invalid %s parameter '%s'" % (type_, name)) from None
    if _lib.ZSTD_isError(bounds.error):
        raise ValueError("invalid %s parameter '%s'" % (type_, name))

    # Error message
    raise ValueError(
        "%s parameter '%s' received an illegal value %d; "
        "the valid range is [%d, %d]"
        % (type_, name, value, bounds.lowerBound, bounds.upperBound)
    )


_PARAMETER_TYPES = {}


def set_parameter_types(c_parameter_type, d_parameter_type):
    """
    Set CompressionParameter and DecompressionParameter types for validity check.

    c_parameter_type
      CompressionParameter IntEnum type object
    d_parameter_type
      DecompressionParameter IntEnum type object
    """
    _PARAMETER_TYPES["compression"] = c_parameter_type
    _PARAMETER_TYPES["decompression"] = d_parameter_type


def _check_int(value):
    if not isinstance(value, int):
        raise TypeError(
            f"'{value.__class__.__name__}' object cannot be interpreted as an integer"
        )
    if value > 2147483647 or value < -2147483648:
        raise OverflowError("Python int too large to convert to C int")


def _clinic_check(fname, args, pos, expected):
    got = type(args[pos])
    if got != expected:
        raise TypeError(
            f"{fname}() argument {pos + 1} must be {expected.__name__}, not {got.__name__}"
        )



from backports.zstd._cffi._common import (
    ZstdError,
    _check_int,
    _clinic_check,
    _ErrorType,
    _ffi,
    _lib,
    _nbytes,
    _new_nonzero,
    _set_zstd_error,
    set_parameter_types,
)
from backports.zstd._cffi.compressor import ZstdCompressor
from backports.zstd._cffi.decompressor import ZstdDecompressor
from backports.zstd._cffi.zstddict import ZstdDict

zstd_version = _ffi.string(_lib.ZSTD_versionString()).decode()
zstd_version_number = _lib.ZSTD_versionNumber()
ZSTD_CLEVEL_DEFAULT = _lib.ZSTD_defaultCLevel()
ZSTD_DStreamOutSize = _lib.ZSTD_DStreamOutSize()

ZSTD_c_compressionLevel = _lib.ZSTD_c_compressionLevel
ZSTD_c_windowLog = _lib.ZSTD_c_windowLog
ZSTD_c_hashLog = _lib.ZSTD_c_hashLog
ZSTD_c_chainLog = _lib.ZSTD_c_chainLog
ZSTD_c_searchLog = _lib.ZSTD_c_searchLog
ZSTD_c_minMatch = _lib.ZSTD_c_minMatch
ZSTD_c_targetLength = _lib.ZSTD_c_targetLength
ZSTD_c_strategy = _lib.ZSTD_c_strategy
ZSTD_c_enableLongDistanceMatching = _lib.ZSTD_c_enableLongDistanceMatching
ZSTD_c_ldmHashLog = _lib.ZSTD_c_ldmHashLog
ZSTD_c_ldmMinMatch = _lib.ZSTD_c_ldmMinMatch
ZSTD_c_ldmBucketSizeLog = _lib.ZSTD_c_ldmBucketSizeLog
ZSTD_c_ldmHashRateLog = _lib.ZSTD_c_ldmHashRateLog
ZSTD_c_contentSizeFlag = _lib.ZSTD_c_contentSizeFlag
ZSTD_c_checksumFlag = _lib.ZSTD_c_checksumFlag
ZSTD_c_dictIDFlag = _lib.ZSTD_c_dictIDFlag
ZSTD_c_nbWorkers = _lib.ZSTD_c_nbWorkers
ZSTD_c_jobSize = _lib.ZSTD_c_jobSize
ZSTD_c_overlapLog = _lib.ZSTD_c_overlapLog

ZSTD_d_windowLogMax = _lib.ZSTD_d_windowLogMax

ZSTD_fast = _lib.ZSTD_fast
ZSTD_dfast = _lib.ZSTD_dfast
ZSTD_greedy = _lib.ZSTD_greedy
ZSTD_lazy = _lib.ZSTD_lazy
ZSTD_lazy2 = _lib.ZSTD_lazy2
ZSTD_btlazy2 = _lib.ZSTD_btlazy2
ZSTD_btopt = _lib.ZSTD_btopt
ZSTD_btultra = _lib.ZSTD_btultra
ZSTD_btultra2 = _lib.ZSTD_btultra2


def get_param_bounds(parameter, is_compress):
    """
    Get CompressionParameter/DecompressionParameter bounds.

    parameter
      The parameter to get bounds.
    is_compress
      True for CompressionParameter, False for DecompressionParameter.
    """
    if is_compress:
        bound = _lib.ZSTD_cParam_getBounds(parameter)
        if _lib.ZSTD_isError(bound.error):
            _set_zstd_error(_ErrorType.ERR_GET_C_BOUNDS, bound.error)
    else:
        bound = _lib.ZSTD_dParam_getBounds(parameter)
        if _lib.ZSTD_isError(bound.error):
            _set_zstd_error(_ErrorType.ERR_GET_D_BOUNDS, bound.error)
    return bound.lowerBound, bound.upperBound


def get_frame_info(frame_buffer):
    """
    Get Zstandard frame infomation from a frame header.

    frame_buffer
      A bytes-like object, containing the header of a Zstandard frame.
    """
    decompressed_size = _lib.ZSTD_getFrameContentSize(
        _ffi.from_buffer(frame_buffer), len(frame_buffer)
    )
    if decompressed_size == _lib.ZSTD_CONTENTSIZE_ERROR:
        raise ZstdError(
            "Error when getting information from the header of "
            "a Zstandard frame. Ensure the frame_buffer argument "
            "starts from the beginning of a frame, and its length "
            "is not less than the frame header (6~18 bytes)."
        )
    dict_id = _lib.ZSTD_getDictID_fromFrame(
        _ffi.from_buffer(frame_buffer), len(frame_buffer)
    )
    if decompressed_size == _lib.ZSTD_CONTENTSIZE_UNKNOWN:
        return None, dict_id
    return decompressed_size, dict_id


def get_frame_size(frame_buffer):
    """
    Get the size of a Zstandard frame, including the header and optional checksum.

    frame_buffer
      A bytes-like object, it should start from the beginning of a frame,
      and contains at least one complete frame.
    """
    frame_size = _lib.ZSTD_findFrameCompressedSize(
        _ffi.from_buffer(frame_buffer), len(frame_buffer)
    )
    if _lib.ZSTD_isError(frame_size):
        raise ZstdError(
            "Error when finding the compressed size of a Zstandard frame. "
            "Ensure the frame_buffer argument starts from the "
            "beginning of a frame, and its length is not less than this "
            "complete frame. Zstd error message: %s."
            % _ffi.string(_lib.ZSTD_getErrorName(frame_size)).decode()
        )
    return frame_size


def _calculate_samples_stats(samples_bytes, samples_sizes):
    chunks_number = len(samples_sizes)
    # Prepare chunk_sizes
    chunk_sizes = _new_nonzero("size_t[]", chunks_number)
    sizes_sum = _nbytes(samples_bytes)
    for i, size in enumerate(samples_sizes):
        try:
            chunk_sizes[i] = size
        except OverflowError:
            sizes_sum = -1
            break
        sizes_sum -= size
    if sizes_sum != 0:
        raise ValueError(
            "The samples size tuple doesn't match the concatenation's size."
        )
    return chunk_sizes, chunks_number


def _clinic_train_dict(*args):
    _clinic_check("train_dict", args, 0, bytes)
    _clinic_check("train_dict", args, 1, tuple)


def train_dict(samples_bytes, samples_sizes, dict_size):
    """
    Train a Zstandard dictionary on sample data.

    samples_bytes
      Concatenation of samples.
    samples_sizes
      Tuple of samples' sizes.
    dict_size
      The size of the dictionary.
    """
    _clinic_train_dict(samples_bytes, samples_sizes, dict_size)

    # Check arguments
    _check_int(dict_size)
    if dict_size <= 0:
        raise ValueError("dict_size argument should be positive number.")

    # Check that the samples are valid and get their sizes
    chunk_sizes, chunks_number = _calculate_samples_stats(samples_bytes, samples_sizes)

    # Allocate dict buffer
    dst_dict_bytes = _new_nonzero("char[]", dict_size)
    if dst_dict_bytes == _ffi.NULL:
        raise MemoryError

    # Train the dictionary
    zstd_ret = _lib.ZDICT_trainFromBuffer(
        dst_dict_bytes,
        dict_size,
        _ffi.from_buffer(samples_bytes),
        chunk_sizes,
        chunks_number,
    )
    if _lib.ZDICT_isError(zstd_ret):
        _set_zstd_error(_ErrorType.ERR_TRAIN_DICT, zstd_ret)

    # Resize dict_buffer
    return _ffi.buffer(dst_dict_bytes)[:zstd_ret]


def _clinic_finalize_dict(*args):
    _clinic_check("finalize_dict", args, 0, bytes)
    _clinic_check("finalize_dict", args, 1, bytes)
    _clinic_check("finalize_dict", args, 2, tuple)


def finalize_dict(
    custom_dict_bytes, samples_bytes, samples_sizes, dict_size, compression_level
):
    """
    Finalize a Zstandard dictionary.

    custom_dict_bytes
      Custom dictionary content.
    samples_bytes
      Concatenation of samples.
    samples_sizes
      Tuple of samples' sizes.
    dict_size
      The size of the dictionary.
    compression_level
      Optimize for a specific Zstandard compression level, 0 means default.
    """
    _clinic_finalize_dict(
        custom_dict_bytes, samples_bytes, samples_sizes, dict_size, compression_level
    )

    # Check arguments
    _check_int(dict_size)
    if dict_size <= 0:
        raise ValueError("dict_size argument should be positive number.")

    # Check that the samples are valid and get their sizes
    chunk_sizes, chunks_number = _calculate_samples_stats(samples_bytes, samples_sizes)

    # Allocate dict buffer
    dst_dict_bytes = _new_nonzero("char[]", dict_size)
    if dst_dict_bytes == _ffi.NULL:
        raise MemoryError

    # Parameters
    params = _new_nonzero("ZDICT_params_t *")
    if params == _ffi.NULL:
        raise MemoryError
    # Optimize for a specific Zstandard compression level, 0 means default.
    params.compressionLevel = compression_level
    # Write log to stderr, 0 = none.
    params.notificationLevel = 0
    # Force dictID value, 0 means auto mode (32-bits random value).
    params.dictID = 0

    # Finalize the dictionary
    zstd_ret = _lib.ZDICT_finalizeDictionary(
        dst_dict_bytes,
        dict_size,
        _ffi.from_buffer(custom_dict_bytes),
        _nbytes(custom_dict_bytes),
        _ffi.from_buffer(samples_bytes),
        chunk_sizes,
        chunks_number,
        params[0],
    )

    # Check Zstandard dict error
    if _lib.ZDICT_isError(zstd_ret):
        _set_zstd_error(_ErrorType.ERR_FINALIZE_DICT, zstd_ret)

    # Resize dict_buffer
    return _ffi.buffer(dst_dict_bytes)[:zstd_ret]



from contextlib import suppress
from threading import Lock

from backports.zstd._cffi._common import (
    ZstdError,
    _check_int,
    _DictionaryType,
    _ffi,
    _lib,
)


class ZstdDict:
    """
    Represents a Zstandard dictionary.

      dict_content
        The content of a Zstandard dictionary as a bytes-like object.
      is_raw
        If true, perform no checks on *dict_content*, useful for some
        advanced cases. Otherwise, check that the content represents
        a Zstandard dictionary created by the zstd library or CLI.

    The dictionary can be used for compression or decompression, and can be shared
    by multiple ZstdCompressor or ZstdDecompressor objects.
    """

    def __init__(self, dict_content, *, is_raw=False):
        # All dictionaries must be at least 8 bytes
        if len(dict_content) < 8:
            raise ValueError(
                "Zstandard dictionary content too short "
                "(must have at least eight bytes)"
            )

        self._d_dict = _ffi.NULL
        self._dict_id = 0
        self._lock = Lock()

        # ZSTD_CDict dict
        self._c_dicts = {}

        self._dict_content = bytes(dict_content)

        # Get dict_id, 0 means "raw content" dictionary.
        self._dict_id = _lib.ZSTD_getDictID_fromDict(
            _ffi.from_buffer(self._dict_content), len(self._dict_content)
        )

        # Check validity for ordinary dictionary
        if not is_raw and self.dict_id == 0:
            raise ValueError("invalid Zstandard dictionary")

    def __repr__(self):
        return "<ZstdDict dict_id=%u dict_size=%d>" % (
            self._dict_id,
            len(self._dict_content),
        )

    def __len__(self):
        return len(self._dict_content)

    @property
    def dict_content(self):
        "The content of a Zstandard dictionary, as a bytes object."
        return self._dict_content

    @property
    def dict_id(self):
        """
        The Zstandard dictionary, an int between 0 and 2**32.

        A non-zero value represents an ordinary Zstandard dictionary,
        conforming to the standardised format.

        A value of zero indicates a 'raw content' dictionary,
        without any restrictions on format or content.
        """
        return self._dict_id

    @property
    def as_digested_dict(self):
        """
        Load as a digested dictionary to compressor.

        Pass this attribute as zstd_dict argument:
        compress(dat, zstd_dict=zd.as_digested_dict)

        1. Some advanced compression parameters of compressor may be overridden
           by parameters of digested dictionary.
        2. ZstdDict has a digested dictionaries cache for each compression level.
           It's faster when loading again a digested dictionary with the same
           compression level.
        3. No need to use this for decompression.
        """
        return self, _DictionaryType.DICT_TYPE_DIGESTED

    @property
    def as_undigested_dict(self):
        """
        Load as an undigested dictionary to compressor.

        Pass this attribute as zstd_dict argument:
        compress(dat, zstd_dict=zd.as_undigested_dict)

        1. The advanced compression parameters of compressor will not be overridden.
        2. Loading an undigested dictionary is costly. If load an undigested dictionary
           multiple times, consider reusing a compressor object.
        3. No need to use this for decompression.
        """
        return self, _DictionaryType.DICT_TYPE_UNDIGESTED

    @property
    def as_prefix(self):
        """
        Load as a prefix to compressor/decompressor.

        Pass this attribute as zstd_dict argument:
        compress(dat, zstd_dict=zd.as_prefix)

        1. Prefix is compatible with long distance matching, while dictionary is not.
        2. It only works for the first frame, then the compressor/decompressor will
           return to no prefix state.
        3. When decompressing, must use the same prefix as when compressing.
        """
        return self, _DictionaryType.DICT_TYPE_PREFIX

    def _get_cdict(self, compression_level):
        with self._lock:
            cdict = self._c_dicts.get(compression_level)
            if cdict is None:
                cdict = _lib.ZSTD_createCDict(
                    self._dict_content, len(self._dict_content), compression_level
                )
                if cdict == _ffi.NULL:
                    raise ZstdError(
                        "Failed to create a ZSTD_CDict instance from "
                        "Zstandard dictionary content."
                    )
                self._c_dicts[compression_level] = cdict

        return cdict

    def _get_ddict(self):
        with self._lock:
            if self._d_dict == _ffi.NULL:
                # Create ZSTD_DDict instance from dictionary content
                self._d_dict = _lib.ZSTD_createDDict(
                    self._dict_content, len(self._dict_content)
                )
                if self._d_dict == _ffi.NULL:
                    raise ZstdError(
                        "Failed to create a ZSTD_DDict instance from "
                        "Zstandard dictionary content."
                    )

            return self._d_dict

    def __del__(self):
        with suppress(AttributeError):
            for level, cdict in self._c_dicts.items():
                _lib.ZSTD_freeCDict(cdict)
                self._c_dicts[level] = _ffi.NULL

        with suppress(AttributeError):
            if self._d_dict != _ffi.NULL:
                _lib.ZSTD_freeDDict(self._d_dict)
                self._d_dict = _ffi.NULL


def _Py_parse_zstd_dict(zd, default_type):
    # Check ZstdDict
    if isinstance(zd, ZstdDict):
        return zd, default_type

    # Check (ZstdDict, type)
    if isinstance(zd, tuple) and len(zd) == 2:
        zd, type_ = zd
        if isinstance(zd, ZstdDict) and isinstance(type_, int):
            _check_int(type_)
            if type_ in (
                _DictionaryType.DICT_TYPE_DIGESTED,
                _DictionaryType.DICT_TYPE_UNDIGESTED,
                _DictionaryType.DICT_TYPE_PREFIX,
            ):
                return zd, type_

    # Wrong type
    raise TypeError("zstd_dict argument should be a ZstdDict object.")



from backports.zstd._cffi._common import _ffi, _lib, _new_nonzero

unable_allocate_msg = "Unable to allocate output buffer."
OUTPUT_BUFFER_MAX_BLOCK_SIZE = 256 * 1024 * 1024
KB = 1024
MB = 1024 * 1024
BUFFER_BLOCK_SIZE = [
    32 * KB,
    64 * KB,
    256 * KB,
    1 * MB,
    4 * MB,
    8 * MB,
    16 * MB,
    16 * MB,
    32 * MB,
    32 * MB,
    32 * MB,
    32 * MB,
    64 * MB,
    64 * MB,
    128 * MB,
    128 * MB,
    OUTPUT_BUFFER_MAX_BLOCK_SIZE,
]


class BlocksOutputBuffer:
    def init_and_grow(self, max_length):
        # get block size
        if 0 <= max_length < BUFFER_BLOCK_SIZE[0]:
            block_size = max_length
        else:
            block_size = BUFFER_BLOCK_SIZE[0]

        # the first block
        b = _new_nonzero("char[]", block_size)
        if b == _ffi.NULL:
            raise MemoryError

        # create the list
        self.list = [b]

        # set variables
        self.allocated = block_size
        self.max_length = max_length

        return block_size, b

    def init_with_size(self, init_size):
        # the first block
        b = _new_nonzero("char[]", init_size)
        if b == _ffi.NULL:
            raise MemoryError

        # create the list
        self.list = [b]

        # set variables
        self.allocated = init_size
        self.max_length = -1

        return init_size, b

    def grow(self, avail_out):
        list_len = len(self.list)

        # ensure no gaps in the data
        if avail_out != 0:
            raise SystemError("avail_out is non-zero in _BlocksOutputBuffer_Grow().")

        # get block size
        if list_len < len(BUFFER_BLOCK_SIZE):
            block_size = BUFFER_BLOCK_SIZE[list_len]
        else:
            block_size = BUFFER_BLOCK_SIZE[-1]

        # check max_length
        if self.max_length >= 0:
            # if (rest == 0), should not grow the buffer.
            rest = self.max_length - self.allocated
            assert rest > 0

            # block_size of the last block
            if block_size > rest:
                block_size = rest

        # create the block
        b = _new_nonzero("char[]", block_size)
        if b == _ffi.NULL:
            raise MemoryError(unable_allocate_msg)
        self.list.append(b)

        # set variables
        self.allocated += block_size

        return block_size, b

    def finish(self, avail_out):
        list_len = len(self.list)

        # fast path for single block
        if (list_len == 1 and avail_out == 0) or (
            list_len == 2 and len(self.list[1]) == avail_out
        ):
            block = self.list[0]
            self.list.clear()
            return bytes(_ffi.buffer(block))

        # final bytes object
        result = _new_nonzero("char[]", self.allocated - avail_out)
        if result == _ffi.NULL:
            raise MemoryError(unable_allocate_msg)

        # memory copy
        if list_len > 0:
            posi = 0

            # blocks except the last one
            for block in self.list[:-1]:
                block_len = len(block)
                _ffi.memmove(result + posi, block, block_len)
                posi += block_len
            # the last block
            block = self.list[-1]
            _ffi.memmove(result + posi, block, len(block) - avail_out)
        else:
            assert len(result) == 0

        self.list.clear()
        return bytes(_ffi.buffer(result))
